# (C) British Crown Copyright 2014, Met Office
#
# This file is part of Biggus.
#
# Biggus is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Biggus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Biggus. If not, see <http://www.gnu.org/licenses/>.
"""Unit tests for `biggus._normalise_axis`."""

import unittest

import mock

from biggus import _normalise_axis


ARRAY = mock.Mock(ndim=9)


class TestNone(unittest.TestCase):
    def test(self):
        self.assertIs(_normalise_axis(None, ARRAY), None)


class TestValid(unittest.TestCase):
    def check(self, argument, expected):
        result = _normalise_axis(argument, ARRAY)
        self.assertEqual(result, expected)
        self.assertIsInstance(result, tuple)

    def test_zero(self):
        self.check(0, (0,))

    def test_one(self):
        self.check(1, (1,))

    def test_single_tuple(self):
        self.check((8,), (8,))

    def test_single_list(self):
        self.check([8], (8,))

    def test_multi_tuple(self):
        self.check((3, 2), (3, 2))

    def test_multi_list(self):
        self.check([4, 1], (4, 1))

    def test_negative(self):
        self.check(-1, (8,))

    def test_negative_tuple(self):
        self.check((-1), (8,))

    def test_mixed(self):
        self.check((-1, 3, -2), (8, 3, 7))


class TestInvalidInt(unittest.TestCase):
    def test_builtin(self):
        with self.assertRaises(TypeError):
            _normalise_axis(open, ARRAY)

    def test_dict(self):
        with self.assertRaises(TypeError):
            _normalise_axis({}, ARRAY)

    def test_float(self):
        with self.assertRaises(TypeError):
            _normalise_axis(2.3, ARRAY)

    def test_str(self):
        with self.assertRaises(TypeError):
            _normalise_axis('23', ARRAY)


class TestInvalidIterable(unittest.TestCase):
    def test_mixed_tuple(self):
        with self.assertRaises(TypeError):
            _normalise_axis((2, '23', 6), ARRAY)

    def test_all_bad_tuple(self):
        with self.assertRaises(TypeError):
            _normalise_axis((2.8, '23', {6: 3}), ARRAY)


if __name__ == '__main__':
    unittest.main()



# (C) British Crown Copyright 2013, Met Office
#
# This file is part of Biggus.
#
# Biggus is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Biggus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Biggus. If not, see <http://www.gnu.org/licenses/>.
import unittest

import numpy as np

import biggus
from biggus.tests import AccessCounter


class TestElementwise(unittest.TestCase):
    def _test_elementwise(self, biggus_op, numpy_op):
        # Sequence of tests, defined as:
        #   1. Original array shape.
        #   2. Sequence of indexing operations to apply.
        tests = [
            [(10, ), []],
            [(30, 40), []],
            [(30, 40), [5]],
            [(500, 30, 40), [slice(3, 6)]],
            [(500, 30, 40), [(slice(None), slice(3, 6))]],
        ]
        axis = 0
        ddof = 0
        for shape, cuts in tests:
            # Define some test data
            size = np.prod(shape)
            raw_data1 = np.linspace(0.0, 1.0, num=size).reshape(shape)
            raw_data2 = np.linspace(0.2, 1.2, num=size).reshape(shape)

            # Check the elementwise operation doesn't actually read any
            # data.
            data1 = AccessCounter(raw_data1)
            data2 = AccessCounter(raw_data2)
            array1 = biggus.NumpyArrayAdapter(data1)
            array2 = biggus.NumpyArrayAdapter(data2)
            op_array = biggus_op(array1, array2)
            self.assertIsInstance(op_array, biggus.Array)
            self.assertTrue((data1.counts == 0).all())
            self.assertTrue((data2.counts == 0).all())

            # Compute the NumPy elementwise operation, and then wrap the
            # result as an array so we can apply biggus-style indexing.
            numpy_op_data = numpy_op(raw_data1, raw_data2)
            numpy_op_array = biggus.NumpyArrayAdapter(numpy_op_data)

            for keys in cuts:
                # Check slicing doesn't actually read any data.
                op_array = op_array[keys]
                self.assertIsInstance(op_array, biggus.Array)
                self.assertTrue((data1.counts == 0).all())
                self.assertTrue((data2.counts == 0).all())
                # Update the NumPy result to match
                numpy_op_array = numpy_op_array[keys]

            # Check the NumPy and biggus numeric values match.
            op_result = op_array.ndarray()
            numpy_result = numpy_op_array.ndarray()
            np.testing.assert_array_equal(op_result, numpy_result)

    def test_add(self):
        self._test_elementwise(biggus.add, np.add)

    def test_sub(self):
        self._test_elementwise(biggus.sub, np.subtract)

    def test_add_nd_array(self):
        # Check that the ElementWise functionality accepts numpy arrays,
        # and the result is as expected.
        r = biggus.add(np.arange(12), np.arange(12))
        self.assertIsInstance(r._array1, biggus.NumpyArrayAdapter)
        self.assertIsInstance(r._array2, biggus.NumpyArrayAdapter)
        self.assertEqual(r.ndarray()[-1], 22)

    def test_add_non_supported_type(self):
        # Check that the ElementWise functionality raises a TypeError
        # if neither an Array or np.ndarray is given
        with self.assertRaises(TypeError):
            biggus.add(range(12), np.arange(12))


if __name__ == '__main__':
    unittest.main()



# (C) British Crown Copyright 2013, Met Office
#
# This file is part of Biggus.
#
# Biggus is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Biggus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Biggus. If not, see <http://www.gnu.org/licenses/>.
import unittest

import numpy as np

import biggus


class TestChaining(unittest.TestCase):
    def test_mean_of_difference(self):
        shape = (3, 4)
        size = np.prod(shape)
        raw_data1 = np.linspace(0.2, 1.0, num=size).reshape(shape)
        raw_data2 = np.linspace(0.3, 1.5, num=size).reshape(shape)
        array1 = biggus.NumpyArrayAdapter(raw_data1)
        array2 = biggus.NumpyArrayAdapter(raw_data2)
        difference = biggus.sub(array2, array1)
        mean_difference = biggus.mean(difference, axis=0)

        # Check the NumPy and biggus numeric values match.
        result = mean_difference.ndarray()
        numpy_result = np.mean(raw_data2 - raw_data1, axis=0)
        np.testing.assert_array_equal(result, numpy_result)

    def test_mean_of_mean(self):
        data = np.arange(24).reshape(3, 4, 2)
        array = biggus.NumpyArrayAdapter(data)
        mean1 = biggus.mean(array, axis=1)
        mean2 = biggus.mean(mean1, axis=-1)
        expected = np.mean(np.mean(data, axis=1), axis=-1)
        result = mean2.ndarray()
        np.testing.assert_array_equal(result, expected)


if __name__ == '__main__':
    unittest.main()



# (C) British Crown Copyright 2012, Met Office
#
# This file is part of Biggus.
#
# Biggus is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Biggus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Biggus. If not, see <http://www.gnu.org/licenses/>.
import unittest

import biggus


class TestArray(unittest.TestCase):
    class _FakeArray(biggus.Array):
        def __init__(self, shape):
            self._shape = shape
        dtype = property(lambda self: None)
        shape = property(lambda self: self._shape)

        def __getitem__(self, keys):
            return None

        def ndarray(self):
            return None

        def masked_array(self):
            return None

    def test_ndim(self):
        shape_ndims = [[(), 0], [(70, 300, 400), 3], [(6,), 1], [(5, 4), 2]]
        for shape, ndim in shape_ndims:
            array = TestArray._FakeArray(shape)
            self.assertEqual(array.ndim, ndim)


if __name__ == '__main__':
    unittest.main()



# (C) British Crown Copyright 2015, Met Office
#
# This file is part of Biggus.
#
# Biggus is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Biggus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Biggus. If not, see <http://www.gnu.org/licenses/>.
"""Unit tests for `biggus.NumpyArrayAdapter`."""

import unittest

import numpy as np
from numpy.testing import assert_array_equal

from biggus import NumpyArrayAdapter, NewAxesArray


class Test___getitem__(unittest.TestCase):
    def test_newaxis(self):
        # Check we have the NewAxesArray.handle_newaxis decorator.
        array = NumpyArrayAdapter(np.arange(24))
        result = array[np.newaxis, :]
        self.assertIsInstance(result, NewAxesArray)
        self.assertEqual(result.shape, (1, 24))


if __name__ == '__main__':
    unittest.main()



# (C) British Crown Copyright 2014, Met Office
#
# This file is part of Biggus.
#
# Biggus is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Biggus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Biggus. If not, see <http://www.gnu.org/licenses/>.
"""Integration tests for swappable engine."""

import unittest

import mock

import biggus


class Test(unittest.TestCase):
    def test(self):
        # If we switch evaluation engine, does it get used?
        array = biggus._Aggregation(biggus.ConstantArray(3, 2), None, None,
                                    None, None, {})
        return_value = (mock.sentinel.result,)
        engine = mock.Mock(**{'ndarrays.return_value': return_value})
        with mock.patch('biggus.engine', engine):
            result = array.ndarray()
        self.assertIs(result, mock.sentinel.result)


if __name__ == '__main__':
    unittest.main()



# (C) British Crown Copyright 2014, Met Office
#
# This file is part of Biggus.
#
# Biggus is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Biggus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Biggus. If not, see <http://www.gnu.org/licenses/>.
"""Unit tests for `biggus.Array`."""

import unittest

import numpy as np

import biggus
from biggus import Array


RESULT_NDARRAY = np.arange(12).reshape(3, 4)


class FakeArray(Array):
    def __init__(self, shape, dtype='f8'):
        self._shape = shape
        self._dtype = np.dtype(dtype)

    @property
    def dtype(self):
        return self._dtype

    @property
    def shape(self):
        return self._shape

    def __getitem__(self, keys):
        pass

    def ndarray(self):
        return RESULT_NDARRAY

    def masked_array(self, keys):
        pass


class Test___array__(unittest.TestCase):
    def test(self):
        array = FakeArray((2, 3))
        result = array.__array__()
        self.assertIs(result, RESULT_NDARRAY)

    def test_dtype(self):
        array = FakeArray((2, 3))
        result = array.__array__('i4')
        np.testing.assert_array_equal(result, RESULT_NDARRAY)
        self.assertEqual(result.dtype, np.dtype('i4'))


class Test_nbytes(unittest.TestCase):
    def _test(self, shape, dtype=np.dtype('f8')):
        array = FakeArray(shape, dtype)
        ndarray = np.empty(shape, dtype)
        self.assertEqual(array.nbytes, ndarray.nbytes)

    def test_type(self):
        self.assertIsInstance(FakeArray((3, 4)).nbytes, int)

    def test_0d(self):
        self._test(())

    def test_1d(self):
        self._test(4)

    def test_nd(self):
        self._test((2, 6, 1, 5))

    def test_dtype(self):
        self._test((2, 6, 1, 5), 'i2')


class Test___str__(unittest.TestCase):
    def _test(self, shape, dtype, expected):
        if not isinstance(shape, tuple):
            shape = (shape,)
        array = FakeArray(shape, dtype)
        self.assertEqual(str(array), expected)

    def test_0d(self):
        self._test((), 'f8',
                   "<Array shape=() dtype=dtype('float64') size=8 B>")

    def test_1d(self):
        self._test(4, 'f8',
                   "<Array shape=(4,) dtype=dtype('float64') size=32 B>")

    def test_nd(self):
        self._test((2, 6, 5), 'f8',
                   "<Array shape=(2, 6, 5) dtype=dtype('float64') size=480 B>")

    def test_1023(self):
        self._test(1023, 'i1',
                   "<Array shape=(1023,) dtype=dtype('int8') size=1023 B>")

    def test_1024(self):
        self._test(1024, 'i1',
                   "<Array shape=(1024,) dtype=dtype('int8') size=1.00 KiB>")

    def test_40000(self):
        self._test(40000, 'i1',
                   "<Array shape=(40000,) dtype=dtype('int8') size=39.06 KiB>")

    def test_999999(self):
        self._test(
            9999999, 'i1',
            "<Array shape=(9999999,) dtype=dtype('int8') size=9.54 MiB>")

    def test_999999999(self):
        self._test(
            9999999999, 'i1',
            "<Array shape=(9999999999,) dtype=dtype('int8') size=9.31 GiB>")

    def test_999999999999(self):
        self._test(
            9999999999999, 'i1',
            "<Array shape=(9999999999999,) dtype=dtype('int8') size=9.09 TiB>")

    def test_999999999999999(self):
        self._test(
            9999999999999999, 'i1',
            "<Array shape=(9999999999999999,) dtype=dtype('int8') "
            "size=9094.95 TiB>")


class Test___hash__(unittest.TestCase):
    def test_unhashable(self):
        array = FakeArray((3, 4), 'f4')
        with self.assertRaises(TypeError):
            hash(array)


class Test_transpose(unittest.TestCase):
    def test_default(self):
        array = FakeArray((2, 3, 4))
        result = array.transpose()
        self.assertIsInstance(result, biggus.TransposedArray)
        self.assertEqual(tuple(result.axes), (2, 1, 0))

    def test_explicit(self):
        array = FakeArray((2, 3, 4))
        result = array.transpose((1, 2, 0))
        self.assertIsInstance(result, biggus.TransposedArray)
        self.assertEqual(result.axes, (1, 2, 0))


if __name__ == '__main__':
    unittest.main()



# (C) British Crown Copyright 2012 - 2013, Met Office
#
# This file is part of Biggus.
#
# Biggus is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Biggus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Biggus. If not, see <http://www.gnu.org/licenses/>.
import unittest

import numpy as np

import biggus


class _TestAdapter(object):
    longMessage = True

    def test_dtype(self):
        dtypes = ['f4', 'i1', 'O', 'm8', '<f4', '>f4', '=f4']
        keys = [(), (5,), (slice(1, 3),)]
        for dtype in dtypes:
            for key in keys:
                array = self.zeros_adapter([10], dtype=dtype, keys=key)
                self.assertEqual(array.dtype, np.dtype(dtype))

    def test_shape_0d(self):
        pairs = [
            [(), ()],
        ]
        for key, shape in pairs:
            array = self.zeros_adapter((), keys=key)
            self.assertEqual(array.shape, shape)

    def test_shape_1d(self):
        pairs = [
            [(), (10,)],
            [(5,), ()],
            [(slice(1, 3),), (2,)],
        ]
        for key, shape in pairs:
            array = self.zeros_adapter([10], keys=key)
            self.assertEqual(array.shape, shape)

    def test_shape_2d(self):
        pairs = [
            [(), (30, 40)],
            [(5,), (40,)],
            [(slice(1, 3),), (2, 40)],
            [(slice(None, None),), (30, 40)],
            [(5, 3), ()],
            [(5, slice(2, 6)), (4,)],
            [(slice(2, 3), slice(2, 6)), (1, 4)],
        ]
        for key, shape in pairs:
            array = self.zeros_adapter((30, 40), keys=key)
            self.assertEqual(array.shape, shape)

    def test_getitem(self):
        # Sequence of tests, defined as:
        #   1. Original array shape,
        #   2. sequence of indexing operations to apply,
        #   3. expected result shape or exception.
        tests = [
            [(30, 40), [], (30, 40)],
            [(30, 40), [5], (40,)],
            [(30, 40), [(5,)], (40,)],
            [(30, 40), [5, 3], ()],
            [(30, 40), [(5,), (4,)], ()],
            [(30, 40), [(slice(None, None), 6)], (30,)],
            [(30, 40), [(slice(None, None), slice(1, 5))], (30, 4)],
            [(30, 40), [(slice(None, None),), 4], (40,)],
            [(30, 40), [5, (slice(None, None),)], (40,)],
            [(30, 40), [(slice(None, 10),)], (10, 40)],
            [(30, 40), [(slice(None, None),)], (30, 40)],
            [(30, 40), [(slice(None, None, -2),)], (15, 40)],
            [(30, 40), [(slice(None, 10),), 5], (40,)],
            [(30, 40), [(slice(None, 10),), (slice(None, 3),)], (3, 40)],
            [(30, 40), [(slice(None, 10),), (slice(None, None, 2),)], (5, 40)],
            [(30, 40), [(slice(5, 10),),
                        (slice(None, None), slice(2, 6))], (5, 4)],
            [(30, 40), [(slice(None, None), slice(2, 6)),
                        (slice(5, 10),)], (5, 4)],
            [(30, 40), [3.5], TypeError],
            [(30, 40), ['foo'], TypeError],
            [(30, 40), [object()], TypeError],
            # Fancy indexing
            [(21, 5, 70, 30, 40), [((1, 5), 0, (2, 5, 10), slice(None, 15))],
                (2, 3, 15, 40)],
            [(21, 5, 2, 70, 30, 40), [(0, (1, 4), 1, (2, 5, 10),
                                       slice(None, 15))], (2, 3, 15, 40)],
            # Boolean indexing
            [(3, 4), [np.array([0, 1, 0], dtype=bool)], (1, 4)],
            [(3, 4), [np.array([1, 0, 1], dtype=bool)], (2, 4)],
            [(3, 4), [np.array([0, 0, 0], dtype=bool)], (0, 4)],
            [(3, 4), [np.array([1, 1, 1], dtype=bool)], (3, 4)],
            [(3, 4), [(slice(None), np.array([1, 0, 1, 1], dtype=bool))],
             (3, 3)],
            [(3, 4), [(slice(None), np.array([0, 1, 0, 0], dtype=bool))],
             (3, 1)],
            [(3, 4), [(slice(None), np.array([1, 1, 1, 1], dtype=bool))],
             (3, 4)],
            [(3, 4), [(slice(None), np.array([0, 0, 0, 0], dtype=bool))],
             (3, 0)],
            # Boolean indexing (too few indices - zero pad)
            [(3, 4), [np.array([1, 1], dtype=bool)], (2, 4)],
            [(3, 4), [(slice(None), np.array([1, 1, 1], dtype=bool))], (3, 3)],
            # Boolean indexing (too many indices)
            [(3, 4), [np.array([1, 1, 1, 0], dtype=bool)], IndexError],
            [(3, 4), [(slice(None), np.array([1, 1, 1, 1, 0], dtype=bool))],
             IndexError],
            # Boolean testing, repeated slicing
            [(3, 4), [(slice(None), slice(None)),
                      np.array([0, 1, 0], dtype=bool)], (1, 4)],
            [(3, 4), [(slice(None), slice(None)),
                      (slice(None), slice(None)),
                      np.array([0, 1, 1], dtype=bool),
                      np.array([1, 0], dtype=bool)], (1, 4)],
        ]
        for src_shape, cuts, target in tests:
            array = self.zeros_adapter(src_shape)
            if isinstance(target, type):
                with self.assertRaises(target):
                    for cut in cuts:
                        array = array.__getitem__(cut)
            else:
                for cut in cuts:
                    array = array.__getitem__(cut)
                    self.assertIsInstance(array, biggus.Array)
                msg = '\nSrc shape: {!r}\nCuts: {!r}'.format(src_shape, cuts)
                self.assertEqual(array.shape, target, msg)
                ndarray = array.ndarray()
                self.assertEqual(ndarray.shape, target, msg)

    def test_ndarray(self):
        tests = [
            [(3,), (), [0, 1, 2]],
            [(3,), (1,), [1]],
            [(3,), (slice(None, None, 2),), [0, 2]],
            [(3, 4), (), [[0, 1, 2, 3], [4, 5, 6, 7], [8, 9, 10, 11]]],
            [(3, 4), (1, ), [4, 5, 6, 7]],
            [(3, 4), (1, 3), 7],
        ]
        for src_shape, src_keys, target in tests:
            array = self.arange_adapter(src_shape, keys=src_keys)
            result = array.ndarray()
            self.assertIsInstance(result, np.ndarray)
            self.assertEqual(array.dtype, result.dtype)
            self.assertEqual(array.shape, result.shape,
                             '\nKeys: {!r}'.format(src_keys))
            np.testing.assert_array_equal(result, target)

    def test_no_ndim(self):
        # The concrete instance should not be need to provide `ndim` for
        # the adapter to construct.
        class Fake(object):
            pass
        ok = Fake()
        ok.shape = (3, 4)
        ok.dtype = 'f'
        array = self.wrap(ok, ())

        no_shape_dtype = Fake()
        with self.assertRaises(AttributeError):
            array = self.wrap(no_shape_dtype, ())

    def zeros_adapter(self, shape, dtype='f', keys=()):
        ndarray = np.zeros(shape, dtype=dtype)
        return self.wrap(ndarray, keys)

    def arange_adapter(self, shape, keys):
        size = reduce(lambda x, y: x * y, shape)
        ndarray = np.arange(size).reshape(shape)
        return self.wrap(ndarray, keys)


class TestNumpyAdapter(_TestAdapter, unittest.TestCase):
    def wrap(self, ndarray, keys):
        return biggus.NumpyArrayAdapter(ndarray, keys)


class TestOrthoAdapter(_TestAdapter, unittest.TestCase):
    class Ortho(object):
        def __init__(self, array):
            self._array = array
            self.shape = array.shape
            self.dtype = array.dtype

        def __getitem__(self, keys):
            result = self._array
            for i, key in reversed(list(enumerate(keys))):
                index = [slice(None)] * i + [key]
                result = result.__getitem__(tuple(index))
            return result

    def wrap(self, ndarray, keys):
        ortho = TestOrthoAdapter.Ortho(ndarray)
        array = biggus.OrthoArrayAdapter(ortho, keys)
        return array


if __name__ == '__main__':
    unittest.main()



# (C) British Crown Copyright 2015, Met Office
#
# This file is part of Biggus.
#
# Biggus is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Biggus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Biggus. If not, see <http://www.gnu.org/licenses/>.
"""Unit tests for `biggus.NewAxesArray`."""

import unittest

import numpy as np
from numpy.testing import assert_array_equal

from biggus import NewAxesArray, ConstantArray, NumpyArrayAdapter


class Test___init__(unittest.TestCase):
    def test_too_few_axes(self):
        in_arr = ConstantArray([3, 2])
        msg = 'must have length 3 but was actually length 2'
        with self.assertRaisesRegexp(ValueError, msg):
            array = NewAxesArray(in_arr, [1, 2])

    def test_too_many_axes(self):
        in_arr = ConstantArray([3, 2])
        msg = 'must have length 3 but was actually length 4'
        with self.assertRaisesRegexp(ValueError, msg):
            array = NewAxesArray(in_arr, [1, 2, 0, 1])

    def test_new_axes_wrong_dtype(self):
        in_arr = ConstantArray([3, 2])
        msg = 'Only positive integer types may be used for new_axes.'
        with self.assertRaisesRegexp(ValueError, msg):
            array = NewAxesArray(in_arr, [1.1, 1, 1])

    def test_new_axes_negative(self):
        in_arr = ConstantArray([1, 2])
        msg = 'Only positive integer types may be used for new_axes.'
        with self.assertRaisesRegexp(ValueError, msg):
            array = NewAxesArray(in_arr, [-1, 0, 1])

    def test_successful_attribute(self):
        in_arr = ConstantArray([3, 1])
        array = NewAxesArray(in_arr, [2, 0, 4])
        self.assertIs(array.array, in_arr)
        self.assertIsInstance(array._new_axes, np.ndarray)
        self.assertEqual(list(array._new_axes), [2, 0, 4])

    def test_0d(self):
        array = NewAxesArray(np.array(0), [3])
        self.assertEqual(list(array._new_axes), [3])

    def test_nested_newaxes(self):
        # Currently it is possible to nest NewAxesArrays. It would be
        # nice if this weren't the case.
        orig = np.empty([3])
        array1 = NewAxesArray(orig, [1, 0])
        array2 = NewAxesArray(array1, [0, 0, 2])
        self.assertEqual(array2.shape, (1, 3, 1, 1))
        self.assertIs(array2.array, array1)


class Test_shape(unittest.TestCase):
    def test_no_new_axes(self):
        in_arr = ConstantArray([3, 1])
        array = NewAxesArray(in_arr, [0, 0, 0])
        self.assertEqual(array.shape, (3, 1))

    def test_many_new_axes(self):
        in_arr = ConstantArray([3, 2])
        array = NewAxesArray(in_arr, [3, 1, 2])
        self.assertEqual(array.shape, (1, 1, 1, 3, 1, 2, 1, 1))

    def test_left_only_new_axes(self):
        in_arr = ConstantArray([3, 2])
        array = NewAxesArray(in_arr, [1, 0, 0])
        self.assertEqual(array.shape, (1, 3, 2))

    def test_right_only_new_axes(self):
        in_arr = ConstantArray([3, 2])
        array = NewAxesArray(in_arr, [0, 0, 2])
        self.assertEqual(array.shape, (3, 2, 1, 1))

    def test_0d(self):
        array = NewAxesArray(np.array(0), [3])
        self.assertEqual(array.shape, (1, 1, 1))


# TODO: Remove once my other PR is in.
class _KeyGen(object):
    def __getitem__(self, keys):
        return keys
key_gen = _KeyGen()


class Test__newaxis_keys(unittest.TestCase):
    def assert_newaxis_keys(self, shape, new_axes, expected):
        in_arr = ConstantArray(shape)
        array = NewAxesArray(in_arr, new_axes)
        keys = array._newaxis_keys()
        self.assertEqual(keys, expected)

    def test_no_new_axes(self):
        self.assert_newaxis_keys([3, 2], [0, 0, 0], key_gen[:, :])

    def test_one_new_axes(self):
        new_ax = np.newaxis
        self.assert_newaxis_keys([3, 2], [1, 1, 1],
                                 key_gen[new_ax, :, new_ax, :, new_ax])

    def test_many_new_axes(self):
        new_ax = np.newaxis
        self.assert_newaxis_keys([3, 2], [3, 2, 2],
                                 key_gen[new_ax, new_ax, new_ax, :,
                                         new_ax, new_ax, :,
                                         new_ax, new_ax])


class Test___getitem__(unittest.TestCase):
    def setUp(self):
        self.array = NewAxesArray(np.arange(24).reshape(4, 3, 2),
                                  [1, 2, 0, 1])
        self.array_3d = NewAxesArray(np.arange(3), [1, 1])

    def test_new_axis_ellipsis_leading(self):
        result = self.array[..., np.newaxis]
        self.assertIsInstance(result, NewAxesArray)
        self.assertEqual(list(result._new_axes), [1, 2, 0, 2])

    def test_new_axis_ellipsis_trailing(self):
        result = self.array[np.newaxis, ...]
        self.assertIsInstance(result, NewAxesArray)
        self.assertEqual(list(result._new_axes), [2, 2, 0, 1])

    def test_new_axis_with_combine(self):
        result = self.array[np.newaxis, :, 0]
        self.assertIsInstance(result, NewAxesArray)
        self.assertEqual(list(result._new_axes), [4, 0, 1])

    def test_index_existing_newaxis(self):
        result = self.array[0, :, 0]
        self.assertIsInstance(result, NewAxesArray)
        self.assertEqual(list(result._new_axes), [0, 1, 0, 1])

    def test_index_existing_newaxis(self):
        result = self.array[0, 0, 0, ..., 0]
        self.assertIsInstance(result, NewAxesArray)
        self.assertEqual(list(result._new_axes), [1, 0, 0])

    def test_index_existing_newaxis(self):
        result = self.array[0, 0, 0, 0, 0, 0, 0]
        self.assertIsInstance(result, NewAxesArray)
        self.assertEqual(list(result._new_axes), [0])

    def test_combine_all_with_newaxis(self):
        result = self.array[0, 0, np.newaxis, 0, 0, np.newaxis, 0, 0, 0]
        self.assertIsInstance(result, NewAxesArray)
        self.assertEqual(list(result._new_axes), [2])

    def test_new_axis_valid_slice(self):
        self.assertEqual(self.array_3d[0:1, ..., 0:1].shape, (1, 3, 1))

    def test_new_axis_invalid_slice(self):
        self.assertEqual(self.array_3d[1:2, ..., 3:1].shape, (0, 3, 0))

    def test_new_axis_valid_index(self):
        self.assertEqual(self.array_3d[0, ..., 0].shape, (3, ))

    def test_new_axis_invalid_index(self):
        with self.assertRaises(IndexError):
            self.array_3d[1]

        with self.assertRaises(IndexError):
            self.array_3d[-2]

    def test_new_axis_tuple_indexing(self):
        self.assertEqual(self.array_3d[(0, 0, 0), ...].shape, (3, 3, 1))

    def test_new_axis_numpy_array_indexing(self):
        msg = "NewAxesArray indexing not yet supported for ndarray keys."
        with self.assertRaisesRegexp(NotImplementedError, msg):
            self.array_3d[np.array([0, 0, 0]), ...]


class Test_ndarray(unittest.TestCase):
    def test_numpy_array(self):
        array = NewAxesArray(np.arange(24), [1, 2])
        self.assertIsInstance(array.ndarray(), np.ndarray)
        assert_array_equal(array.masked_array(),
                           np.arange(24).reshape(1, 24, 1, 1))


class Test_masked_array(unittest.TestCase):
    def test_numpy_array(self):
        array = NewAxesArray(np.arange(24), [1, 2])
        self.assertIsInstance(array.masked_array(), np.ma.MaskedArray)
        assert_array_equal(array.masked_array(),
                           np.arange(24).reshape(1, 24, 1, 1))


if __name__ == '__main__':
    unittest.main()



# (C) British Crown Copyright 2014, Met Office
#
# This file is part of Biggus.
#
# Biggus is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Biggus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Biggus. If not, see <http://www.gnu.org/licenses/>.
"""Unit tests for `biggus.size`."""

import unittest

from biggus import size


class FakeArray(object):
    def __init__(self, nbytes):
        self.nbytes = nbytes


class Test(unittest.TestCase):
    def _test(self, shape, dtype, expected):
        if not isinstance(shape, tuple):
            shape = (shape,)
        array = FakeArray(shape, dtype)
        self.assertEqual(str(array), expected)

    def test_bytes(self):
        self.assertEqual(size(FakeArray(8)), '8 B')

    def test_1023(self):
        self.assertEqual(size(FakeArray(1023)), '1023 B')

    def test_1024(self):
        self.assertEqual(size(FakeArray(1024)), '1.00 KiB')

    def test_kib(self):
        self.assertEqual(size(FakeArray(4e4)), '39.06 KiB')

    def test_mib(self):
        self.assertEqual(size(FakeArray(4e6)), '3.81 MiB')

    def test_gib(self):
        self.assertEqual(size(FakeArray(4e9)), '3.73 GiB')

    def test_tib(self):
        self.assertEqual(size(FakeArray(4e12)), '3.64 TiB')

    def test_eib(self):
        self.assertEqual(size(FakeArray(4e15)), '3637.98 TiB')


if __name__ == '__main__':
    unittest.main()



import numpy as np


class AccessCounter(object):
    """
    Something that acts like a NumPy ndarray, but which records how
    many times each element has been read.

    """
    def __init__(self, ndarray):
        self._ndarray = ndarray
        self.counts = np.zeros(ndarray.shape)

    @property
    def dtype(self):
        return self._ndarray.dtype

    @property
    def ndim(self):
        return self._ndarray.ndim

    @property
    def shape(self):
        return self._ndarray.shape

    def __array__(self):
        return self._ndarray

    def __getitem__(self, keys):
        self.counts[keys] += 1
        return self._ndarray[keys]

    def unique_counts(self):
        return set(np.unique(self.counts))


class _KeyGen(object):
    """Gives you the content of keys when indexing."""
    def __getitem__(self, keys):
        return keys

#: An object that can be indexed to return a usable key.
key_gen = _KeyGen()



# (C) British Crown Copyright 2014, Met Office
#
# This file is part of Biggus.
#
# Biggus is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Biggus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Biggus. If not, see <http://www.gnu.org/licenses/>.
"""Unit tests for `biggus.count`."""

import numpy as np
import numpy.ma as ma
import unittest

import biggus
from biggus import count


class TestInvalidAxis(unittest.TestCase):
    def setUp(self):
        self.array = biggus.NumpyArrayAdapter(np.arange(12))

    def test_none(self):
        with self.assertRaises(biggus.AxisSupportError):
            count(self.array)

    def test_too_large(self):
        with self.assertRaises(ValueError):
            count(self.array, axis=1)

    def test_too_small(self):
        with self.assertRaises(ValueError):
            count(self.array, axis=-2)

    def test_multiple(self):
        array = biggus.NumpyArrayAdapter(np.arange(12).reshape(3, 4))
        with self.assertRaises(biggus.AxisSupportError):
            count(array, axis=(0, 1))


class TestAggregationDtype(unittest.TestCase):
    def _check(self, source):
        array = biggus.NumpyArrayAdapter(np.arange(2, dtype=source))
        agg = count(array, axis=0)
        self.assertEqual(agg.dtype, np.dtype('i'))

    def test_int_to_float(self):
        dtypes = [np.int8, np.int16, np.int32, np.int]
        for dtype in dtypes:
            self._check(dtype)

    def test_bool_to_float(self):
        self._check(np.bool)

    def test_floats(self):
        dtypes = [np.float16, np.float32, np.float]
        for dtype in dtypes:
            self._check(dtype)

    def test_complex(self):
        self._check(np.complex)


class TestNumpyArrayAdapter(unittest.TestCase):
    def setUp(self):
        self.data = np.arange(12)

    def _check(self, data, dtype=None, shape=None):
        data = np.asarray(data, dtype=dtype)
        if shape is not None:
            data = data.reshape(shape)
        array = biggus.NumpyArrayAdapter(data)
        result = count(array, axis=0).ndarray()
        expected = np.ones(data.shape[1:]) * data.shape[0]
        np.testing.assert_array_equal(result, expected)

    def test_flat_int(self):
        self._check(self.data)

    def test_multi_int(self):
        self._check(self.data, shape=(3, 4))

    def test_flat_float(self):
        self._check(self.data, dtype=np.float)

    def test_multi_float(self):
        self._check(self.data, dtype=np.float, shape=(3, 4))


class TestNumpyArrayAdapterMasked(unittest.TestCase):
    def _check(self, data):
        array = biggus.NumpyArrayAdapter(data)
        result = count(array, axis=0).masked_array()
        expected = ma.asarray(ma.count(data, axis=0))
        np.testing.assert_array_equal(result.filled(), expected)
        np.testing.assert_array_equal(result.mask, expected.mask)

    def test_no_mask_flat(self):
        for dtype in [np.int, np.float]:
            data = ma.arange(12, dtype=dtype)
            self._check(data)

    def test_no_mask_multi(self):
        for dtype in [np.int, np.float]:
            data = ma.arange(12, dtype=dtype).reshape(3, 4)
            self._check(data)

    def test_flat(self):
        for dtype in [np.int, np.float]:
            data = ma.arange(12, dtype=dtype)
            data[::2] = ma.masked
            self._check(data)

            data.mask = ma.nomask
            data[1::2] = ma.masked
            self._check(data)

    def test_multi(self):
        for dtype in [np.int, np.float]:
            data = ma.arange(12, dtype=dtype)
            data[::2] = ma.masked
            self._check(data.reshape(3, 4))

            data = ma.arange(12, dtype=dtype)
            data[1::2] = ma.masked
            self._check(data.reshape(3, 4))

            data = ma.arange(12, dtype=dtype).reshape(3, 4)
            data[::2] = ma.masked
            self._check(data)

            data = ma.arange(12, dtype=dtype).reshape(3, 4)
            data[1::2] = ma.masked
            self._check(data)


if __name__ == '__main__':
    unittest.main()



from functools import wraps

import numpy as np

import biggus


def _wrap_shape(np_func):
    @wraps(np_func)
    def func(shape):
        return _RandomShape(np_func, shape)
    return func

def _wrap_sizes(np_func):
    @wraps(np_func)
    def func(*sizes):
        return _RandomSizes(np_func, sizes)
    return func


rand = _wrap_sizes(np.random.rand)
randn = _wrap_sizes(np.random.randn)

random = _wrap_shape(np.random.random)


class _RandomShape(biggus.Array):
    def __init__(self, func, shape):
        self.func = func
        self._shape = shape

    @property
    def shape(self):
        return self._shape

    @property
    def dtype(self):
        return np.dtype(float)

    def __getitem__(self, keys):
        shape = biggus._sliced_shape(self.shape, keys)
        return _RandomShape(self.func, shape)

    def ndarray(self):
        return self.func(self.shape)

    def masked_array(self):
        return np.ma.asarray(self.ndarray())


class _RandomSizes(biggus.Array):
    def __init__(self, np_func, shape):
        self.np_func = np_func
        self._shape = shape

    @property
    def shape(self):
        return self._shape

    @property
    def dtype(self):
        return np.dtype(float)

    def __getitem__(self, keys):
        shape = biggus._sliced_shape(self.shape, keys)
        return _RandomSizes(self.func, shape)

    def ndarray(self):
        return self.func(self.shape)

    def masked_array(self):
        return np.ma.asarray(self.ndarray())



# (C) British Crown Copyright 2014, Met Office
#
# This file is part of Biggus.
#
# Biggus is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Biggus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Biggus. If not, see <http://www.gnu.org/licenses/>.
"""Unit tests for `biggus._all_slices`."""

import unittest

import mock

import numpy as np

import biggus
from biggus import _all_slices


class Test__all_slices(unittest.TestCase):
    def _small_array(self):
        shape = (5, 928, 744)
        data = np.arange(np.prod(shape), dtype=np.float32).reshape(shape)
        array = biggus.NumpyArrayAdapter(data)
        return array

    def test_min(self):
        array = self._small_array()
        slices = _all_slices(array)
        expected = [[slice(0, 3, None), slice(3, 5, None)],
                    (slice(None, None, None),), (slice(None, None, None),)]
        self.assertEqual(slices, expected)

if __name__ == '__main__':
    unittest.main()



# (C) British Crown Copyright 2014, Met Office
#
# This file is part of Biggus.
#
# Biggus is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Biggus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Biggus. If not, see <http://www.gnu.org/licenses/>.
"""Integration tests for biggus."""



# (C) British Crown Copyright 2013 - 2014, Met Office
#
# This file is part of Biggus.
#
# Biggus is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Biggus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Biggus. If not, see <http://www.gnu.org/licenses/>.
import unittest

import numpy as np
import numpy.ma
import numpy.testing

import biggus
from biggus.tests import AccessCounter


class TestAggregation(unittest.TestCase):
    def _test_aggregation(self, biggus_op, numpy_op, **kwargs):
        # Sequence of tests, defined as:
        #   1. Original array shape.
        #   2. Sequence of indexing operations to apply.
        tests = [
            [(10, ), []],
            [(30, 40), []],
            [(30, 40), [5]],
            [(500, 30, 40), [slice(3, 6)]],
            [(500, 30, 40), [(slice(None), slice(3, 6))]],
        ]
        for shape, cuts in tests:
            # Define some test data
            size = np.prod(shape)
            raw_data = np.linspace(0, 1, num=size).reshape(shape)

            for axis in range(len(shape)):
                # Check the aggregation operation doesn't actually read any
                # data.
                data = AccessCounter(raw_data)
                array = biggus.NumpyArrayAdapter(data)
                op_array = biggus_op(array, axis=axis, **kwargs)
                self.assertIsInstance(op_array, biggus.Array)
                self.assertTrue((data.counts == 0).all())

                # Compute the NumPy aggregation, and then wrap the result as
                # an array so we can apply biggus-style indexing.
                numpy_op_data = numpy_op(raw_data, axis=axis, **kwargs)
                numpy_op_array = biggus.NumpyArrayAdapter(numpy_op_data)

                for keys in cuts:
                    # Check slicing doesn't actually read any data.
                    op_array = op_array[keys]
                    self.assertIsInstance(op_array, biggus.Array)
                    self.assertTrue((data.counts == 0).all())
                    # Update the NumPy result to match
                    numpy_op_array = numpy_op_array[keys]

                # Check resolving `op_array` to a NumPy array only reads
                # each relevant source value once.
                op_result = op_array.ndarray()
                self.assertTrue((data.counts <= 1).all())

                # Check the NumPy and biggus numeric values match.
                numpy_result = numpy_op_array.ndarray()
                np.testing.assert_array_almost_equal(op_result, numpy_result)

    def test_mean(self):
        self._test_aggregation(biggus.mean, np.mean)

    def test_std(self):
        self._test_aggregation(biggus.std, np.std, ddof=0)
        self._test_aggregation(biggus.std, np.std, ddof=1)

    def test_var(self):
        self._test_aggregation(biggus.var, np.var, ddof=0)
        self._test_aggregation(biggus.var, np.var, ddof=1)

    def test_mean_nd_array(self):
        r = biggus.mean(np.arange(12), axis=0)
        self.assertIsInstance(r._array, biggus.NumpyArrayAdapter)
        self.assertEqual(r.ndarray(), 5.5)

    def test_add_non_supported_type(self):
        # Check that the Aggregation raises a TypeError
        # if neither an Array or np.ndarray is given
        msg = "list' object has no attribute 'ndim"
        with self.assertRaisesRegexp(AttributeError, msg):
            biggus.mean(range(10), axis=0)


class TestMdtol(unittest.TestCase):
    def setUp(self):
        self.data = np.ma.arange(12).reshape(3, 4)
        self.data[2, 1:] = np.ma.masked

    def _test_mean_with_mdtol(self, data, axis, numpy_result, mdtol=None):
        data = AccessCounter(data)
        array = biggus.NumpyArrayAdapter(data)

        # Perform aggregation.
        if mdtol is None:
            # Allow testing of default when mdtol is None.
            biggus_aggregation = biggus.mean(array, axis=axis)
        else:
            biggus_aggregation = biggus.mean(array, axis=axis, mdtol=mdtol)

        # Check the aggregation operation doesn't actually read any data.
        self.assertTrue((data.counts == 0).all())

        # Check results.
        biggus_result = biggus_aggregation.masked_array()
        # Check resolving `op_array` to a NumPy array only reads
        # each relevant source value once.
        self.assertTrue((data.counts <= 1).all())
        numpy_mask = np.ma.getmaskarray(numpy_result)
        biggus_mask = np.ma.getmaskarray(biggus_result)
        np.testing.assert_array_equal(biggus_mask, numpy_mask)
        np.testing.assert_array_equal(biggus_result[~biggus_mask].data,
                                      numpy_result[~numpy_mask].data)

    def test_mean_mdtol_default(self):
        axis = 0
        expected = np.ma.mean(self.data, axis)
        self._test_mean_with_mdtol(self.data, axis, expected)

    def test_mean_mdtol_one(self):
        axis = 0
        mdtol = 1
        expected = np.ma.mean(self.data, axis)
        self._test_mean_with_mdtol(self.data, axis, expected, mdtol)

    def test_mean_mdtol_zero(self):
        axis = 0
        mdtol = 0
        expected = np.ma.mean(self.data, axis)
        expected.mask = [False, True, True, True]
        self._test_mean_with_mdtol(self.data, axis, expected, mdtol)

    def test_mean_mdtol_fraction_below_axis_zero(self):
        axis = 0
        mdtol = 0.32
        expected = np.ma.mean(self.data, axis)
        expected.mask = [False, True, True, True]
        self._test_mean_with_mdtol(self.data, axis, expected, mdtol)

    def test_mean_mdtol_fraction_above_axis_zero(self):
        axis = 0
        mdtol = 0.34
        expected = np.ma.mean(self.data, axis)
        self._test_mean_with_mdtol(self.data, axis, expected, mdtol)

    def test_mean_mdtol_fraction_below_axis_one(self):
        axis = 1
        mdtol = 0.74
        expected = np.ma.mean(self.data, axis)
        expected.mask = [False, False, True]
        self._test_mean_with_mdtol(self.data, axis, expected, mdtol)

    def test_mean_mdtol_fraction_above_axis_one(self):
        axis = 1
        mdtol = 0.76
        expected = np.ma.mean(self.data, axis)
        self._test_mean_with_mdtol(self.data, axis, expected, mdtol)


class TestFlow(unittest.TestCase):
    def _test_flow(self, axis):
        data = np.arange(3 * 4 * 5, dtype='f4').reshape(3, 4, 5)
        array = biggus.NumpyArrayAdapter(data)
        mean = biggus.mean(array, axis=axis)
        engine = biggus.AllThreadedEngine()
        chunk_size = biggus.MAX_CHUNK_SIZE
        try:
            # Artificially constrain the chunk size to eight bytes to
            # ensure biggus is stepping across axes in the correct
            # order.
            biggus.MAX_CHUNK_SIZE = 8
            op_result, = engine.ndarrays(mean)
        finally:
            biggus.MAX_CHUNK_SIZE = chunk_size
        np_result = np.mean(data, axis=axis)
        np.testing.assert_array_almost_equal(op_result, np_result)

    def test_axis_0(self):
        self._test_flow(0)

    def test_axis_1(self):
        self._test_flow(1)

    def test_axis_2(self):
        self._test_flow(2)


if __name__ == '__main__':
    unittest.main()



# (C) British Crown Copyright 2014, Met Office
#
# This file is part of Biggus.
#
# Biggus is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Biggus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Biggus. If not, see <http://www.gnu.org/licenses/>.
"""Unit tests for `biggus._ArrayAdapter`."""

import unittest

import numpy as np
from numpy.testing import assert_array_equal
from numpy.lib.stride_tricks import as_strided


from biggus import BroadcastArray, NumpyArrayAdapter


class Test___init__(unittest.TestCase):
    def test_array(self):
        orig = np.empty([1, 3, 5])
        a = BroadcastArray(orig, {})
        self.assertIs(a.array, orig)

    def test_invalid_broadcast_axis(self):
        msg = 'Axis -1 out of range \[0, 5\)'
        with self.assertRaisesRegexp(ValueError, msg):
            a = BroadcastArray(np.empty([1, 3, 1, 5, 1]), {-1: 10})

    def test_invalid_broadcast_length(self):
        msg = 'Axis length must be positive. Got -1.'
        with self.assertRaisesRegexp(ValueError, msg):
            a = BroadcastArray(np.empty([1, 3, 1, 5, 1]), {0: -1})

    def test_broadcasting_existing_non_len1_dimension(self):
        msg = 'Attempted to broadcast axis 0 which is of length 3.'
        with self.assertRaisesRegexp(ValueError, msg):
            a = BroadcastArray(np.empty([3]), {0: 5})

    def test_nested_broadcast_avoidance(self):
        orig = np.empty([1, 3, 1, 5, 1])
        a = BroadcastArray(orig, {0: 10, 4: 100})
        b = BroadcastArray(a, {0: 5, 2: 15})
        self.assertIs(b.array, orig)
        self.assertEqual(a._broadcast_dict, {0: 10, 4: 100})
        self.assertEqual(b._broadcast_dict, {0: 5, 2: 15, 4: 100})


class Test_shape(unittest.TestCase):
    def test_broadcast_shape(self):
        a = BroadcastArray(np.empty([1, 3, 1, 5, 1]), {0: 10, 2: 0, 4: 15})
        self.assertEqual(a.shape, (10, 3, 0, 5, 15))


class Test___getitem__(unittest.TestCase):
    def test_nothing_done(self):
        orig = np.empty([1, 3, 1, 5, 1])
        a = BroadcastArray(orig, {0: 10, 2: 0, 4: 15})
        result = a[...]
        self.assertIsInstance(result, BroadcastArray)
        self.assertIs(result.array, orig)
        self.assertEqual(result.shape, (10, 3, 0, 5, 15))

    def test_index_exising_broadcast(self):
        orig = np.empty([1, 3, 1, 5, 1])
        a = BroadcastArray(orig, {0: 10, 2: 0, 4: 15})
        result = a[:-1]
        self.assertIs(result.array, orig)
        self.assertEqual(result.shape, (9, 3, 0, 5, 15))

    def test_index_contained_array_dimension(self):
        orig = np.empty([1, 3, 1, 5, 1])
        a = BroadcastArray(orig, {0: 10, 2: 0, 4: 15})
        result = a[:, -1]
        assert_array_equal(result.array, orig[:, -1])
        self.assertEqual(result.shape, (10, 0, 5, 15))


class Test_ndarray(unittest.TestCase):
    def test_indexed(self):
        orig = np.empty([1, 3, 1, 5, 1], dtype='>i4')
        a = BroadcastArray(orig, {0: 10, 2: 0, 4: 15})
        result = a[:, -1, ::2, ::2]
        expected = as_strided(orig, shape=(10, 0, 3, 15),
                              strides=(0, 0, 4, 0))
        assert_array_equal(result.ndarray(), expected)


class Test_masked_array(unittest.TestCase):
    def test_simple(self):
        orig = np.ma.masked_array([[1], [2], [3]],
                                  mask=[[1], [0], [1]])
        array = BroadcastArray(orig, {1: 2})
        result = array.masked_array()
        expected = np.ma.masked_array([[1, 1], [2, 2], [3, 3]],
                                      mask=[[1, 1], [0, 0], [1, 1]])
        assert_array_equal(result.mask, expected.mask)
        assert_array_equal(result.data, expected.data)

    def test_indexed(self):
        orig = np.ma.masked_array([[1], [2], [3]], mask=[[1], [0], [1]])
        a = BroadcastArray(orig, {1: 2})
        result = a[0:2, :-1].masked_array()
        expected = np.ma.masked_array([[1], [2]], mask=[[1], [0]])
        assert_array_equal(result.mask, expected.mask)
        assert_array_equal(result.data, expected.data)


class Test_broadcast_numpy_array(unittest.TestCase):
    def test_simple_broadcast(self):
        a = np.arange(3, dtype='>i4').reshape([3, 1])
        result = BroadcastArray.broadcast_numpy_array(a, {1: 4})
        expected = np.array([[0, 0, 0, 0],
                             [1, 1, 1, 1],
                             [2, 2, 2, 2]])
        assert_array_equal(result, expected)
        self.assertEqual(result.strides, (4, 0))


if __name__ == '__main__':
    unittest.main()



# (C) British Crown Copyright 2013, Met Office
#
# This file is part of Biggus.
#
# Biggus is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Biggus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Biggus. If not, see <http://www.gnu.org/licenses/>.
"""
Perform a PEP8 conformance test of the Biggus code base.

"""
import os
import unittest

import pep8

import biggus


class TestCodeFormat(unittest.TestCase):
    """Test code conformance."""

    def test_pep8_conformance(self):
        # Tests the biggus code base against the "pep8" tool.
        #
        # Users can add their own excluded files (should files exist in the
        # local directory which is not in the repository) by adding a
        # ".pep8_test_exclude.txt" file in the same directory as this test.
        # The file should be a line separated list of filenames/directories
        # as can be passed to the "pep8" tool's exclude list.

        pep8style = pep8.StyleGuide(quiet=False)
        pep8style.options.exclude.extend([])

        # Allow users to add their own exclude list.
        test_root = os.path.abspath(__file__)
        extra_exclude_file = os.path.join(os.path.dirname(test_root),
                                          '.pep8_test_exclude.txt')
        if os.path.exists(extra_exclude_file):
            with open(extra_exclude_file, 'r') as fhandle:
                extra_exclude = [line.strip()
                                 for line in fhandle if line.strip()]
            pep8style.options.exclude.extend(extra_exclude)

        root = os.path.abspath(biggus.__file__)
        result = pep8style.check_files([os.path.dirname(root)])
        self.assertEqual(result.total_errors, 0, "Found code syntax "
                                                 "errors (and warnings).")


if __name__ == '__main__':
    unittest.main()



# (C) British Crown Copyright 2012 - 2015, Met Office
#
# This file is part of Biggus.
#
# Biggus is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Biggus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Biggus. If not, see <http://www.gnu.org/licenses/>.
import unittest

import numpy as np

import biggus


class TestStack(unittest.TestCase):
    def test_dtype(self):
        dtype = np.dtype('f4')
        item = biggus.NumpyArrayAdapter(np.empty(6, dtype=dtype))
        stack = np.array([item], dtype='O')
        array = biggus.ArrayStack(stack)
        self.assertEqual(array.dtype, dtype)

    def test_shape_and_getitem(self):
        # Sequence of tests, defined as:
        #   1. Stack shape,
        #   2. item shape,
        #   3. sequence of indexing operations to apply,
        #   4. expected result shape or exception.
        tests = [
            [(6, 70), (30, 40), [], (6, 70, 30, 40)],
            [(6, 70), (30, 40), [5], (70, 30, 40,)],
            [(6, 70), (30, 40), [(5,)], (70, 30, 40,)],
            [(6, 70), (30, 40), [5, 3], (30, 40)],
            [(6, 70), (30, 40), [(5, 3)], (30, 40)],
            [(6, 70), (30, 40), [5, 3, 2], (40,)],
            [(6, 70), (30, 40), [(5, 3, 2)], (40,)],
            [(6, 70), (30, 40), [5, 3, 2, 1], ()],
            [(6, 70), (30, 40), [(5, 3, 2, 1)], ()],
            [(6, 70), (30, 40), [5, (3, 2), 1], ()],
            [(6, 70), (30, 40), [(slice(None, None), 6)], (6, 30, 40)],
            [(6, 70), (30, 40), [(slice(None, None), np.newaxis, 6)],
             (6, 1, 30, 40)],
            [(6, 70), (30, 40), [(slice(None, None), slice(1, 5))],
             (6, 4, 30, 40)],
            [(6, 70), (30, 40), [(slice(None, None),), 4], (70, 30, 40,)],
            [(6, 70), (30, 40), [5, (slice(None, None),)], (70, 30, 40,)],
            [(6, 70), (30, 40), [(slice(None, 10),)], (6, 70, 30, 40)],
            [(6, 70), (30, 40), [(slice(None, 10),), 5], (70, 30, 40,)],
            [(6, 70), (30, 40), [(slice(None, 10),), (slice(None, 3),)],
             (3, 70, 30, 40)],
            [(6, 70), (30, 40), [(slice(None, 10),), (slice(None, None, 2),)],
             (3, 70, 30, 40)],
            [(6, 70), (30, 40), [(slice(5, 10),),
                                 (slice(None, None), slice(2, 6))],
             (1, 4, 30, 40)],
            [(6, 70), (30, 40), [(slice(None, None), slice(2, 6)),
                                 (slice(5, 10),)], (1, 4, 30, 40)],
            [(6, 70), (30, 40), [3.5], ValueError],
            [(6, 70), (30, 40), ['foo'], ValueError],
            [(6, 70), (30, 40), [object()], ValueError],
        ]
        dtype = np.dtype('f4')
        for stack_shape, item_shape, cuts, target in tests:

            def make_array(*n):
                concrete = np.empty(item_shape, dtype)
                array = biggus.NumpyArrayAdapter(concrete)
                return array

            stack = np.empty(stack_shape, dtype='O')
            for index in np.ndindex(stack.shape):
                stack[index] = make_array()
            array = biggus.ArrayStack(stack)
            if isinstance(target, type):
                with self.assertRaises(target):
                    for cut in cuts:
                        array = array.__getitem__(cut)
            else:
                for cut in cuts:
                    array = array.__getitem__(cut)
                    self.assertIsInstance(array, biggus.Array)
                self.assertEqual(array.shape, target,
                                 '\nCuts: {!r}'.format(cuts))

    def test_ndarray(self):
        # Sequence of tests, defined as:
        #   1. Stack shape,
        #   2. item shape,
        #   3. expected result.
        tests = [
            [(1,), (3,), np.arange(3).reshape(1, 3)],
            [(1,), (3, 4), np.arange(12).reshape(1, 3, 4)],
            [(6,), (3, 4), np.arange(72).reshape(6, 3, 4)],
            [(6, 70), (3, 4), np.arange(5040).reshape(6, 70, 3, 4)],
        ]
        for stack_shape, item_shape, target in tests:
            stack = np.empty(stack_shape, dtype='O')
            item_size = np.array(item_shape).prod()
            for index in np.ndindex(stack.shape):
                start = np.ravel_multi_index(index, stack_shape) * item_size
                concrete = np.arange(item_size).reshape(item_shape)
                concrete += start
                array = biggus.NumpyArrayAdapter(concrete)
                stack[index] = array
            array = biggus.ArrayStack(stack)
            result = array.ndarray()
            self.assertIsInstance(result, np.ndarray)
            self.assertEqual(array.dtype, result.dtype)
            self.assertEqual(array.shape, result.shape)
            np.testing.assert_array_equal(result, target)


if __name__ == '__main__':
    unittest.main()



# (C) British Crown Copyright 2014, Met Office
#
# This file is part of Biggus.
#
# Biggus is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Biggus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Biggus. If not, see <http://www.gnu.org/licenses/>.
"""Integration tests for aggregations."""

import unittest

import numpy as np

import biggus


class Test(unittest.TestCase):
    def test_var_masked(self):
        data = biggus.ConstantArray((10, 400, 720), dtype=np.float32)
        v = biggus.var(data, axis=1)
        result = v.masked_array()
        self.assertIsInstance(result, np.ma.MaskedArray)
        self.assertEqual(result.shape, (10, 720))
        self.assertTrue(np.all(result == 0))
        self.assertTrue(np.all(result.mask == 0))


if __name__ == '__main__':
    unittest.main()



# (C) British Crown Copyright 2014, Met Office
#
# This file is part of Biggus.
#
# Biggus is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Biggus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Biggus. If not, see <http://www.gnu.org/licenses/>.
"""Unit tests for `biggus.ConstantArray`."""

import unittest

import mock
import numpy as np

from biggus import ConstantArray


class Test___init__(unittest.TestCase):
    def test_shape_scalar(self):
        shape = 5
        array = ConstantArray(shape)
        self.assertEqual(array.shape, (shape,))

    def test_shape_tuple(self):
        shape = (30, 10, 20)
        array = ConstantArray(shape)
        self.assertEqual(array.shape, shape)

    def test_shape_tuple(self):
        shape = (30, 10, 20)
        array = ConstantArray(shape)
        self.assertEqual(array.shape, shape)

    def test_shape_scalar_like_int(self):
        array = ConstantArray('34')
        self.assertEqual(array.shape, (34,))

    def test_shape_tuple_like_int(self):
        array = ConstantArray(('34', '93'))
        self.assertEqual(array.shape, (34, 93))

    def test_shape_invalid_scalar(self):
        with self.assertRaises(ValueError):
            array = ConstantArray('foo')

    def test_shape_invalid_tuple(self):
        with self.assertRaises(ValueError):
            array = ConstantArray(('foo', 'bar'))

    def test_value(self):
        value = 6
        array = ConstantArray(3, value)
        self.assertEqual(array.value, value)

    def test_dtype(self):
        dtype = 'i2'
        array = ConstantArray((), dtype=dtype)
        self.assertIs(array.dtype, np.dtype(dtype))

    def test_dtype_default(self):
        array = ConstantArray(())
        self.assertEqual(array.dtype, np.dtype('f8'))

    def test_dtype_default_integer(self):
        array = ConstantArray((), 42)
        self.assertEqual(array.dtype, np.dtype(np.int_))


class Test___getitem__(unittest.TestCase):
    def test_indexing_slice(self):
        shape = (30, 10, 20)
        array = ConstantArray(shape)
        result = array[:5]
        data = np.zeros(shape)[:5]
        self.assertEqual(result.shape, data.shape)

    def test_newaxis(self):
        array = ConstantArray([2, 3])
        result = array[:5, np.newaxis]
        self.assertEqual(result.shape, (2, 1, 3))


class Test_masked_array(unittest.TestCase):
    def test_masked(self):
        shape = (3, 4, 5)
        array = ConstantArray(shape)
        result = array.masked_array()
        self.assertTrue(np.ma.isMaskedArray(result))

    def test_values(self):
        shape = (3, 4, 5)
        array = ConstantArray(shape)
        result = array.masked_array()
        np.testing.assert_array_equal(result, np.ma.zeros(shape))

    def test_dtype(self):
        shape = (3, 4, 5)
        array = ConstantArray(shape, dtype='i4')
        result = array.masked_array()
        self.assertEqual(result.dtype, np.dtype('i4'))

    def test_dtype_default(self):
        shape = (3, 4, 5)
        array = ConstantArray(shape)
        result = array.masked_array()
        self.assertEqual(result.dtype, np.dtype('f8'))


class Test_ndarray(unittest.TestCase):
    def test_values(self):
        shape = (3, 4, 5)
        value = 81
        array = ConstantArray(shape, value)
        result = array.ndarray()
        expected = np.empty(shape)
        expected[()] = value
        np.testing.assert_array_equal(result, expected)

    def test_values_default(self):
        shape = (3, 4, 5)
        array = ConstantArray(shape)
        result = array.ndarray()
        np.testing.assert_array_equal(result, np.zeros(shape))

    def test_dtype(self):
        shape = (3, 4, 5)
        array = ConstantArray(shape, dtype='i4')
        result = array.ndarray()
        self.assertEqual(result.dtype, np.dtype('i4'))

    def test_dtype_default(self):
        shape = (3, 4, 5)
        array = ConstantArray(shape)
        result = array.ndarray()
        self.assertEqual(result.dtype, np.dtype('f8'))


if __name__ == '__main__':
    unittest.main()



# (C) British Crown Copyright 2013, Met Office
#
# This file is part of Biggus.
#
# Biggus is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Biggus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Biggus. If not, see <http://www.gnu.org/licenses/>.
import unittest

import numpy as np

import biggus
from biggus.tests import AccessCounter


class TestNdarray(unittest.TestCase):
    def assert_counts(self, counts, expected_counts):
        self.assertEqual(np.unique(counts), expected_counts)

    def test_dual_aggregation(self):
        # Check the aggregation operations don't actually read any data.
        shape = (500, 30, 40)
        size = np.prod(shape)
        raw_data = np.linspace(0, 1, num=size).reshape(shape)
        counter = AccessCounter(raw_data)
        array = biggus.NumpyArrayAdapter(counter)
        mean_array = biggus.mean(array, axis=0)
        std_array = biggus.std(array, axis=0)
        self.assertIsInstance(mean_array, biggus.Array)
        self.assertIsInstance(std_array, biggus.Array)
        self.assertTrue((counter.counts == 0).all())

        mean, std_dev = biggus.ndarrays([mean_array, std_array])

        # Was the source data read just once?
        self.assert_counts(counter.counts, [1])

    def test_mean_of_difference(self):
        # MEAN(A - B)
        shape = (500, 30, 40)
        size = np.prod(shape)
        raw_data = np.linspace(0, 1, num=size).reshape(shape)
        data = AccessCounter(raw_data * 3)
        a_array = biggus.NumpyArrayAdapter(data)
        data = AccessCounter(raw_data)
        b_array = biggus.NumpyArrayAdapter(data)

        mean_array = biggus.mean(biggus.sub(a_array, b_array), axis=0)

        mean = mean_array.ndarray()
        np.testing.assert_array_almost_equal(mean,
                                             np.mean(raw_data * 2, axis=0))

    def test_sd_and_mean_of_difference(self):
        # MEAN(A - B) and SD(A - B)
        shape = (500, 30, 40)
        size = np.prod(shape)
        raw_data = np.linspace(0, 1, num=size).reshape(shape)
        a_counter = AccessCounter(raw_data * 3)
        a_array = biggus.NumpyArrayAdapter(a_counter)
        b_counter = AccessCounter(raw_data)
        b_array = biggus.NumpyArrayAdapter(b_counter)

        sub_array = biggus.sub(a_array, b_array)
        mean_array = biggus.mean(sub_array, axis=0)
        std_array = biggus.std(sub_array, axis=0)
        mean, std = biggus.ndarrays([mean_array, std_array])

        # Are the resulting numbers equivalent?
        np.testing.assert_array_almost_equal(mean,
                                             np.mean(raw_data * 2, axis=0))
        np.testing.assert_array_almost_equal(std,
                                             np.std(raw_data * 2, axis=0))
        # Was the source data read just once?
        self.assert_counts(a_counter.counts, [1])
        self.assert_counts(b_counter.counts, [1])

    def test_dual_mean_of_difference(self):
        # MEAN(B - A) and MEAN(C - A)
        shape = (500, 30, 40)
        size = np.prod(shape)
        raw_data = np.linspace(0, 1, num=size).reshape(shape)
        a_counter = AccessCounter(raw_data)
        a_array = biggus.NumpyArrayAdapter(a_counter)
        b_counter = AccessCounter(raw_data * 3)
        b_array = biggus.NumpyArrayAdapter(b_counter)
        c_counter = AccessCounter(raw_data * 5)
        c_array = biggus.NumpyArrayAdapter(c_counter)

        b_sub_a_array = biggus.sub(b_array, a_array)
        mean_b_sub_a_array = biggus.mean(b_sub_a_array, axis=0)
        c_sub_a_array = biggus.sub(c_array, a_array)
        mean_c_sub_a_array = biggus.mean(c_sub_a_array, axis=0)

        mean_b_sub_a, mean_c_sub_a = biggus.ndarrays([mean_b_sub_a_array,
                                                      mean_c_sub_a_array])

        # Are the resulting numbers equivalent?
        np.testing.assert_array_almost_equal(mean_b_sub_a,
                                             np.mean(raw_data * 2, axis=0))
        np.testing.assert_array_almost_equal(mean_c_sub_a,
                                             np.mean(raw_data * 4, axis=0))
        # Was the source data read just once?
        self.assert_counts(a_counter.counts, [1])
        self.assert_counts(b_counter.counts, [1])
        self.assert_counts(c_counter.counts, [1])

    def test_mean_of_a_and_mean_of_difference(self):
        # MEAN(A) and MEAN(A - B)
        shape = (500, 30, 40)
        size = np.prod(shape)
        raw_data = np.linspace(0, 1, num=size).reshape(shape)
        a_counter = AccessCounter(raw_data * 3)
        a_array = biggus.NumpyArrayAdapter(a_counter)
        b_counter = AccessCounter(raw_data)
        b_array = biggus.NumpyArrayAdapter(b_counter)

        sub_array = biggus.sub(a_array, b_array)
        mean_a_array = biggus.mean(a_array, axis=0)
        mean_sub_array = biggus.mean(sub_array, axis=0)
        mean_a, mean_sub = biggus.ndarrays([mean_a_array, mean_sub_array])

        # Are the resulting numbers equivalent?
        np.testing.assert_array_almost_equal(mean_a,
                                             np.mean(raw_data * 3, axis=0))
        np.testing.assert_array_almost_equal(mean_sub,
                                             np.mean(raw_data * 2, axis=0))

        # Was the source data read the minimal number of times?
        self.assert_counts(a_counter.counts, [1])
        self.assert_counts(b_counter.counts, [1])

    def test_means_across_different_axes(self):
        # MEAN(A, axis=0) and MEAN(A, axis=1)
        shape = (500, 30, 40)
        size = np.prod(shape)
        raw_data = np.linspace(0, 1, num=size).reshape(shape)
        a_counter = AccessCounter(raw_data * 3)
        a_array = biggus.NumpyArrayAdapter(a_counter)

        mean_0_array = biggus.mean(a_array, axis=0)
        mean_1_array = biggus.mean(a_array, axis=1)
        mean_0, mean_1 = biggus.ndarrays([mean_0_array, mean_1_array])

        # Are the resulting numbers equivalent?
        np.testing.assert_array_almost_equal(mean_0,
                                             np.mean(raw_data * 3, axis=0))
        np.testing.assert_array_almost_equal(mean_1,
                                             np.mean(raw_data * 3, axis=1))

        # Was the source data read the minimal number of times?
        self.assert_counts(a_counter.counts, [1])


if __name__ == '__main__':
    unittest.main()



# (C) British Crown Copyright 2014, Met Office
#
# This file is part of Biggus.
#
# Biggus is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Biggus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Biggus. If not, see <http://www.gnu.org/licenses/>.
"""Unit tests for `biggus.LinearMosaic`."""

import unittest

import mock
import numpy as np

from biggus import Array, LinearMosaic, ConstantArray


class Test___init___invalid(unittest.TestCase):
    def test_not_arrays(self):
        class BadArray(object):
            dtype = 'f'
            fill_value = 9
            ndim = 1
            shape = (4,)
        bad_arrays = [BadArray()]
        with self.assertRaisesRegexp(ValueError, 'subclass'):
            LinearMosaic(bad_arrays, 0)


def fake_array(fill_value, dtype=np.dtype('f4')):
    return mock.Mock(shape=(3, 4), dtype=dtype, fill_value=fill_value,
                     ndim=2, spec=Array)


class Test___init___fill_values(unittest.TestCase):
    def test_nan_nan(self):
        array1 = fake_array(np.nan)
        array2 = fake_array(np.nan)
        mosaic = LinearMosaic(np.array([array1, array2]), 0)
        self.assertTrue(np.isnan(mosaic.fill_value))

    def test_nan_number(self):
        array1 = fake_array(np.nan)
        array2 = fake_array(42)
        mosaic = LinearMosaic(np.array([array1, array2]), 0)
        self.assertEqual(mosaic.fill_value, 1e+20)

    def test_number_nan(self):
        array1 = fake_array(42)
        array2 = fake_array(np.nan)
        mosaic = LinearMosaic(np.array([array1, array2]), 0)
        self.assertEqual(mosaic.fill_value, 1e+20)

    def test_number_number(self):
        array1 = fake_array(42)
        array2 = fake_array(42)
        mosaic = LinearMosaic(np.array([array1, array2]), 0)
        self.assertEqual(mosaic.fill_value, 42)

    def test_number_other_number(self):
        array1 = fake_array(42)
        array2 = fake_array(43)
        mosaic = LinearMosaic(np.array([array1, array2]), 0)
        self.assertEqual(mosaic.fill_value, 1e+20)

    def test_matching_strings(self):
        array1 = fake_array('foo', np.dtype('S3'))
        array2 = fake_array('foo', np.dtype('S3'))
        mosaic = LinearMosaic(np.array([array1, array2]), 0)
        self.assertEqual(mosaic.fill_value, 'foo')

    def test_different_strings(self):
        array1 = fake_array('foo', np.dtype('S3'))
        array2 = fake_array('bar', np.dtype('S3'))
        mosaic = LinearMosaic(np.array([array1, array2]), 0)
        self.assertEqual(mosaic.fill_value, 'N/A')


class Test___getitem__(unittest.TestCase):
    # Note, these are not a complete set of unit tests.
    # Currently they only handle the newaxis checking.
    # There are more tests in biggus.tests.test_linear_mosaic.
    def setUp(self):
        self.a1 = ConstantArray([2, 3])
        self.a2 = ConstantArray([4, 3])
        self.a = LinearMosaic([self.a1, self.a2], axis=0)

    def test_newaxis_leading(self):
        self.assertEqual(self.a[np.newaxis].shape, (1, 6, 3))

    def test_newaxis_trailing(self):
        self.assertEqual(self.a[..., np.newaxis].shape, (6, 3, 1))


if __name__ == '__main__':
    unittest.main()



# (C) British Crown Copyright 2014, Met Office
#
# This file is part of Biggus.
#
# Biggus is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Biggus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Biggus. If not, see <http://www.gnu.org/licenses/>.
"""Unit tests for `biggus.ArrayStack`."""

import unittest

import mock
import numpy as np

from biggus import Array, ArrayStack, NumpyArrayAdapter, ConstantArray


class Test___init___invalid(unittest.TestCase):
    def test_not_arrays(self):
        class BadArray(object):
            dtype = 'f'
            fill_value = 9
            shape = ()
        bad_arrays = np.array([BadArray()])
        with self.assertRaisesRegexp(ValueError, 'subclass'):
            ArrayStack(bad_arrays)


def fake_array(fill_value, dtype=np.dtype('f4')):
    return mock.Mock(shape=mock.sentinel.SHAPE, dtype=dtype,
                     fill_value=fill_value, spec=Array)


class Test___init___fill_values(unittest.TestCase):
    def test_nan_nan(self):
        array1 = fake_array(np.nan)
        array2 = fake_array(np.nan)
        stack = ArrayStack(np.array([array1, array2]))
        self.assertTrue(np.isnan(stack.fill_value))

    def test_nan_number(self):
        array1 = fake_array(np.nan)
        array2 = fake_array(42)
        stack = ArrayStack(np.array([array1, array2]))
        self.assertEqual(stack.fill_value, 1e+20)

    def test_number_nan(self):
        array1 = fake_array(42)
        array2 = fake_array(np.nan)
        stack = ArrayStack(np.array([array1, array2]))
        self.assertEqual(stack.fill_value, 1e+20)

    def test_number_number(self):
        array1 = fake_array(42)
        array2 = fake_array(42)
        stack = ArrayStack(np.array([array1, array2]))
        self.assertEqual(stack.fill_value, 42)

    def test_number_other_number(self):
        array1 = fake_array(42)
        array2 = fake_array(43)
        stack = ArrayStack(np.array([array1, array2]))
        self.assertEqual(stack.fill_value, 1e+20)

    def test_matching_strings(self):
        array1 = fake_array('foo', np.dtype('S3'))
        array2 = fake_array('foo', np.dtype('S3'))
        stack = ArrayStack(np.array([array1, array2]))
        self.assertEqual(stack.fill_value, 'foo')

    def test_different_strings(self):
        array1 = fake_array('foo', np.dtype('S3'))
        array2 = fake_array('bar', np.dtype('S3'))
        stack = ArrayStack(np.array([array1, array2]))
        self.assertEqual(stack.fill_value, 'N/A')


class Test_multidim_array_stack(unittest.TestCase):
    def setUp(self):
        self.arrays = [ConstantArray((), i) for i in range(6)]

    def test_stack_order_c_numpy_array_t1(self):
        # 1D stack of arrays shape (6,)
        res = ArrayStack.multidim_array_stack(self.arrays, (3, 2), order='C')
        arr = np.array([i for i in range(6)])
        target = np.reshape(arr, (3, 2), order='C')
        self.assertTrue(np.array_equal(res.ndarray(), target))

    def test_stack_order_c_numpy_array_t2(self):
        # 1D stack of arrays shape (6,)
        # alternate shape
        res = ArrayStack.multidim_array_stack(self.arrays, (2, 3), order='C')
        arr = np.array([i for i in range(6)])
        target = np.reshape(arr, (2, 3), order='C')
        self.assertTrue(np.array_equal(res.ndarray(), target))

    def test_stack_order_c_multidim(self):
        # 1D stack of 6 arrays each (4, 5)
        arrays = [NumpyArrayAdapter(np.arange(20).reshape(4, 5)) for
                  i in range(6)]
        res = ArrayStack.multidim_array_stack(arrays, (2, 3), order='C')
        arr = np.arange(20).reshape(4, 5) * np.ones((6, 4, 5))
        target = np.reshape(arr, (2, 3, 4, 5), order='C')
        self.assertTrue(np.array_equal(res.ndarray(), target))

    def test_stack_order_default(self):
        # 1D stack of arrays shape (6,)
        # Ensure that the default index ordering corresponds to C.
        res = ArrayStack.multidim_array_stack(self.arrays, (3, 2))
        arr = np.array([i for i in range(6)])
        target = np.reshape(arr, (3, 2), order='C')
        self.assertTrue(np.array_equal(res.ndarray(), target))

    def test_stack_order_fortran_t1(self):
        # Fortran index ordering
        # 1D stack of arrays shape (6,)
        res = ArrayStack.multidim_array_stack(self.arrays, (3, 2), order='F')
        arr = np.array([i for i in range(6)])
        target = np.reshape(arr, (3, 2), order='F')
        self.assertTrue(np.array_equal(res.ndarray(), target))

    def test_stack_order_fortran_t2(self):
        # Fortran index ordering
        # 1D stack of arrays shape (6,)
        # alternate shape
        res = ArrayStack.multidim_array_stack(self.arrays, (2, 3), order='F')
        arr = np.array([i for i in range(6)])
        target = np.reshape(arr, (2, 3), order='F')
        self.assertTrue(np.array_equal(res.ndarray(), target))

    def test_incompatible_shape(self):
        # 1D stack of arrays shape (6,)
        # Specifying a stack shape that is not feasible.
        msg = 'total size of new array must be unchanged'
        with self.assertRaisesRegexp(ValueError, msg):
            ArrayStack.multidim_array_stack(self.arrays, (3, 1), order='C')

    def test_multidim_stack_multidim(self):
        # Multidim stack of arrays shape (4, 6)
        arrays = [[ConstantArray((), i) for i in range(6)] for i in range(4)]
        msg = 'multidimensional stacks not yet supported'
        with self.assertRaisesRegexp(ValueError, msg):
            ArrayStack.multidim_array_stack(arrays, (3, 2, 4))

    def test_order_incorrect_order(self):
        # Specifying an unknown order.
        array1 = fake_array(0)
        array2 = fake_array(0)
        with self.assertRaisesRegexp(TypeError, 'order not understood'):
            ArrayStack.multidim_array_stack([array1, array2], (1, 2),
                                            order='random')


class Test___getitem__(unittest.TestCase):
    # Note, these are not a complete set of unit tests.
    # Currently they only handle the newaxis checking.
    # There are more tests in biggus.tests.test_stack.
    def setUp(self):
        self.a1 = ConstantArray([4, 3])
        self.a2 = ConstantArray([4, 3])
        self.a = ArrayStack([self.a1, self.a2])

    def test_newaxis_leading(self):
        self.assertEqual(self.a[np.newaxis].shape, (1, 2, 4, 3))

    def test_newaxis_trailing(self):
        self.assertEqual(self.a[..., np.newaxis].shape, (2, 4, 3, 1))


if __name__ == '__main__':
    unittest.main()



# (C) British Crown Copyright 2013, Met Office
#
# This file is part of Biggus.
#
# Biggus is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Biggus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Biggus. If not, see <http://www.gnu.org/licenses/>.
import unittest

import numpy as np

import biggus


class TestLinearMosaic(unittest.TestCase):
    def _tile(self, shape, dtype='f4'):
        data = np.arange(np.product(shape), dtype=dtype).reshape(shape)
        return biggus.NumpyArrayAdapter(data)

    def test_init(self):
        # Simple
        tile3x4 = self._tile((3, 4))
        mosaic = biggus.LinearMosaic(tile3x4, 0)
        mosaic = biggus.LinearMosaic([tile3x4], 0)

        # 2-dimensional tile array => error
        with self.assertRaises(ValueError):
            mosaic = biggus.LinearMosaic([[tile3x4], [tile3x4]], 0)

        # Different axis values
        mosaic = biggus.LinearMosaic([tile3x4, tile3x4], 0)
        mosaic = biggus.LinearMosaic([tile3x4, tile3x4], 1)
        with self.assertRaises(ValueError):
            mosaic = biggus.LinearMosaic([tile3x4, tile3x4], -1)
        with self.assertRaises(ValueError):
            mosaic = biggus.LinearMosaic([tile3x4, tile3x4], 2)

        # Tile shapes
        tile3x5 = self._tile((3, 5))
        with self.assertRaises(ValueError):
            mosaic = biggus.LinearMosaic([tile3x4, tile3x5], 0)
        mosaic = biggus.LinearMosaic([tile3x4, tile3x5], 1)

        tile5x4 = self._tile((5, 4))
        mosaic = biggus.LinearMosaic([tile3x4, tile5x4], 0)
        with self.assertRaises(ValueError):
            mosaic = biggus.LinearMosaic([tile3x4, tile5x4], 1)

        # dtypes
        tile3x4_i4 = self._tile((3, 4), 'i4')
        with self.assertRaises(ValueError):
            mosaic = biggus.LinearMosaic([tile3x4, tile3x4_i4], 0)

    def test_ndim(self):
        tile3x4 = self._tile((3, 4))
        mosaic = biggus.LinearMosaic(tile3x4, 0)
        self.assertEqual(mosaic.ndim, 2)

        mosaic = biggus.LinearMosaic([tile3x4, tile3x4], 0)
        self.assertEqual(mosaic.ndim, 2)

        mosaic = biggus.LinearMosaic([tile3x4, tile3x4], 1)
        self.assertEqual(mosaic.ndim, 2)

        tile3 = self._tile((3,))
        mosaic = biggus.LinearMosaic(tile3, 0)
        self.assertEqual(mosaic.ndim, 1)

        tile3x4x5 = self._tile((3, 4, 5))
        mosaic = biggus.LinearMosaic(tile3x4x5, 0)
        self.assertEqual(mosaic.ndim, 3)

    def test_dtype(self):
        for spec in ('f4', 'i2'):
            dtype = np.dtype(spec)
            mosaic = biggus.LinearMosaic(self._tile(6, dtype), 0)
            self.assertEqual(mosaic.dtype, dtype)

    def test_shape(self):
        tile3x4 = self._tile((3, 4))
        mosaic = biggus.LinearMosaic(tile3x4, 0)
        self.assertEqual(mosaic.shape, (3, 4))

        tile2x4 = self._tile((2, 4))
        mosaic = biggus.LinearMosaic([tile3x4, tile2x4], 0)
        self.assertEqual(mosaic.shape, (5, 4))

        tile3x5 = self._tile((3, 5))
        mosaic = biggus.LinearMosaic([tile3x4, tile3x5], 1)
        self.assertEqual(mosaic.shape, (3, 9))

    def test_getitem(self):
        tile3x4 = self._tile((3, 4))
        # NB. This is also testing ndarray(), so perhaps a different
        # name would be more accurate.
        mosaic = biggus.LinearMosaic(tile3x4, 0)

        result = mosaic[1].ndarray()
        target = [4, 5, 6, 7]
        np.testing.assert_array_equal(result, target)

        result = mosaic[:, 2].ndarray()
        target = [2, 6, 10]
        np.testing.assert_array_equal(result, target)

        result = mosaic[2, 1].ndarray()
        target = 9
        np.testing.assert_array_equal(result, target)

        mosaic = biggus.LinearMosaic([tile3x4, tile3x4], 1)

        result = mosaic[1].ndarray()
        target = [4, 5, 6, 7, 4, 5, 6, 7]
        np.testing.assert_array_equal(result, target)

        result = mosaic[1:2].ndarray()
        target = [[4, 5, 6, 7, 4, 5, 6, 7]]
        np.testing.assert_array_equal(result, target)

        result = mosaic[0:2].ndarray()
        target = [[0, 1, 2, 3, 0, 1, 2, 3],
                  [4, 5, 6, 7, 4, 5, 6, 7]]
        np.testing.assert_array_equal(result, target)

        result = mosaic[2:0:-1].ndarray()
        target = [[8, 9, 10, 11, 8, 9, 10, 11],
                  [4, 5, 6, 7, 4, 5, 6, 7]]
        np.testing.assert_array_equal(result, target)

        result = mosaic[:, 5].ndarray()
        target = [1, 5, 9]
        np.testing.assert_array_equal(result, target)

        result = mosaic[:, 1::3].ndarray()
        target = [[1, 0, 3],
                  [5, 4, 7],
                  [9, 8, 11]]
        np.testing.assert_array_equal(result, target)

        result = mosaic[::-2].ndarray()
        target = [[8, 9, 10, 11, 8, 9, 10, 11],
                  [0, 1, 2, 3, 0, 1, 2, 3]]
        np.testing.assert_array_equal(result, target)

        result = mosaic[::-2, 1::3].ndarray()
        target = [[9, 8, 11],
                  [1, 0, 3]]
        np.testing.assert_array_equal(result, target)

        result = mosaic[(0, 0, 2, 1), 1::3].ndarray()
        target = [[1, 0, 3],
                  [1, 0, 3],
                  [9, 8, 11],
                  [5, 4, 7]]
        np.testing.assert_array_equal(result, target)

        result = mosaic[:, (3, 6, 1)].ndarray()
        target = [[3, 2, 1],
                  [7, 6, 5],
                  [11, 10, 9]]
        np.testing.assert_array_equal(result, target)

        result = mosaic[1, 6].ndarray()
        target = 6
        np.testing.assert_array_equal(result, target)

        with self.assertRaises(TypeError):
            result = mosaic['foo'].ndarray()

    def test_getitem_adjust_axis(self):
        tile_1 = self._tile((3, 4, 5))
        tile_2 = self._tile((3, 4, 2))
        mosaic = biggus.LinearMosaic([tile_1, tile_2], 2)
        result = mosaic[1, :, :]
        self.assertEqual(result.shape, (4, 7))
        result = mosaic[:, 1, :]
        self.assertEqual(result.shape, (3, 7))

    def test_getitem_adjust_axis_2(self):
        tile_1 = self._tile((3, 4, 5))
        tile_2 = self._tile((3, 4, 2))
        mosaic = biggus.LinearMosaic([tile_1, tile_2], 2)
        result = mosaic[1, 1, :]
        self.assertEqual(result.shape, (7,))

    def test_ndarray(self):
        tile3x4 = self._tile((3, 4))
        mosaic = biggus.LinearMosaic(tile3x4, 0)
        result = mosaic.ndarray()
        target = [[0, 1, 2, 3],
                  [4, 5, 6, 7],
                  [8, 9, 10, 11]]
        np.testing.assert_array_equal(result, target)

        tile2x4 = self._tile((2, 4))
        mosaic = biggus.LinearMosaic([tile3x4, tile2x4], 0)
        result = mosaic.ndarray()
        target = [[0, 1, 2, 3],
                  [4, 5, 6, 7],
                  [8, 9, 10, 11],
                  [0, 1, 2, 3],
                  [4, 5, 6, 7]]
        np.testing.assert_array_equal(result, target)

        tile3x5 = self._tile((3, 5))
        mosaic = biggus.LinearMosaic([tile3x4, tile3x5], 1)
        result = mosaic.ndarray()
        target = [[0, 1, 2, 3, 0, 1, 2, 3, 4],
                  [4, 5, 6, 7, 5, 6, 7, 8, 9],
                  [8, 9, 10, 11, 10, 11, 12, 13, 14]]
        np.testing.assert_array_equal(result, target)


if __name__ == '__main__':
    unittest.main()



# (C) British Crown Copyright 2014, Met Office
#
# This file is part of Biggus.
#
# Biggus is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Biggus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Biggus. If not, see <http://www.gnu.org/licenses/>.
"""Unit tests for `biggus.max`."""

import numpy as np
import numpy.ma as ma
import unittest

import biggus
import biggus.tests.unit._aggregation_test_framework as test_framework


class Operator(object):
    @property
    def biggus_operator(self):
        return biggus.max

    @property
    def numpy_operator(self):
        return np.max

    @property
    def numpy_masked_operator(self):
        return ma.max


class TestInvalidAxis(Operator, test_framework.InvalidAxis, unittest.TestCase):
    pass


class TestAggregationDtype(
        Operator, test_framework.AggregationDtype, unittest.TestCase):
    pass


class TestNumpyArrayAdapter(
        Operator, test_framework.NumpyArrayAdapter, unittest.TestCase):
    pass


class TestNumpyArrayAdapterMasked(
        Operator, test_framework.NumpyArrayAdapterMasked, unittest.TestCase):
    pass


if __name__ == '__main__':
    unittest.main()



# (C) British Crown Copyright 2014, Met Office
#
# This file is part of Biggus.
#
# Biggus is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Biggus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Biggus. If not, see <http://www.gnu.org/licenses/>.
"""Unit tests for `biggus.zeros`."""

import unittest

import mock
import numpy as np

from biggus import zeros


class Test(unittest.TestCase):
    def test_dtype(self):
        dtype = 'i2'
        array = zeros((), dtype)
        self.assertIs(array.dtype, np.dtype(dtype))

    def test_dtype_default(self):
        array = zeros(())
        self.assertEqual(array.dtype, np.dtype('f8'))

    def test_shape(self):
        shape = (70, 768, 1024)
        array = zeros(shape)
        self.assertEqual(array.shape, shape)

    def test_value(self):
        array = zeros(())
        self.assertEqual(array.value, 0)


if __name__ == '__main__':
    unittest.main()



# (C) British Crown Copyright 2015, Met Office
#
# This file is part of Biggus.
#
# Biggus is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Biggus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Biggus. If not, see <http://www.gnu.org/licenses/>.
"""Unit tests for `biggus.ArrayContainer`."""

import unittest

import numpy as np
from numpy.testing import assert_array_equal

from biggus import ArrayContainer, NumpyArrayAdapter


class Test_ndarray(unittest.TestCase):
    def test_biggus_array(self):
        array = ArrayContainer(NumpyArrayAdapter(np.arange(24)))
        self.assertIsInstance(array.ndarray(), np.ndarray)
        assert_array_equal(array.ndarray(),
                           np.arange(24))

    def test_numpy_array(self):
        array = ArrayContainer(np.arange(24))
        self.assertIsInstance(array.ndarray(), np.ndarray)
        assert_array_equal(array.ndarray(),
                           np.arange(24))


class Test_masked_array(unittest.TestCase):
    def test_biggus_array(self):
        array = ArrayContainer(NumpyArrayAdapter(np.arange(24)))
        self.assertIsInstance(array.masked_array(), np.ma.MaskedArray)
        assert_array_equal(array.masked_array(), np.arange(24))

    def test_numpy_array(self):
        array = ArrayContainer(np.arange(24))
        self.assertIsInstance(array.masked_array(), np.ma.MaskedArray)
        assert_array_equal(array.masked_array(),
                           np.arange(24))


if __name__ == '__main__':
    unittest.main()



# (C) British Crown Copyright 2015, Met Office
#
# This file is part of Biggus.
#
# Biggus is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Biggus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Biggus. If not, see <http://www.gnu.org/licenses/>.
"""Unit tests for `biggus._sliced_shape`."""

import unittest

import numpy as np

from biggus.tests import key_gen
from biggus import _sliced_shape


class Test__sliced_shape(unittest.TestCase):
    def assertSliceShape(self, shape, keys, expected, not_numpy=False):
        self.assertEqual(_sliced_shape(shape, keys), expected)
        if not not_numpy:
            np_actual = np.empty(shape)[keys].shape
            # Will only fail if numpy is not doing what the test anticipates.
            self.assertEqual(np_actual, expected)

    def test_all_scalar(self):
        keys = key_gen[0, 2, 1]
        self.assertSliceShape([4, 5, 6], keys, ())

    def test_all_sliced(self):
        keys = key_gen[:, :, :]
        self.assertSliceShape([4, 5, 6], keys, (4, 5, 6))

    def test_negative_indexing(self):
        keys = key_gen[:-1, -1:, ::-2]
        self.assertSliceShape([4, 5, 6], keys, (3, 1, 3))

    def test_crazy_sizes(self):
        keys = key_gen[1:10, 1:15, 11:10]
        self.assertSliceShape([4, 5, 6], keys, (3, 4, 0))

    def test_new_axis(self):
        keys = key_gen[np.newaxis, :, np.newaxis, :, 0]
        self.assertSliceShape([4, 5, 6], keys, (1, 4, 1, 5))

    def test_numpy_array_indexing_single(self):
        keys = key_gen[np.arange(2), :, 0]
        self.assertSliceShape([4, 5, 6], keys, (2, 5))

    def test_numpy_array_indexing_double(self):
        keys = key_gen[np.arange(2), :, np.arange(1, 4)]
        self.assertSliceShape([4, 5, 6], keys, (2, 5, 3), not_numpy=True)

    def test_tuple_indexing(self):
        keys = key_gen[(1, 2), :, 0]
        self.assertSliceShape([4, 5, 6], keys, (2, 5))

    def test_invalid_object_indexing(self):
        keys = key_gen[np.nan]
        msg = 'Invalid indexing object "nan"'
        with self.assertRaisesRegexp(ValueError, msg):
            _sliced_shape([4, 5, 6], keys)

    def test_invalid_object_indexing_float(self):
        # A float is a valid indexing object in numpy.
        keys = key_gen[1.2]
        msg = 'Invalid indexing object "1.2"'
        with self.assertRaisesRegexp(ValueError, msg):
            _sliced_shape([4, 5, 6], keys)

    def test_numpy_bool_indexing(self):
        keys = key_gen[0, :, np.arange(6) > 2]
        # The numpy bool indexing appears completely confused for >1d cases.
        self.assertSliceShape([4, 5, 6], keys, (5, 3), not_numpy=True)

    def test_all_sliced_ellipsis(self):
        keys = key_gen[:, :, :, ...]
        self.assertEqual(_sliced_shape([3, 2, 1], keys),
                         (3, 2, 1))


if __name__ == '__main__':
    unittest.main()
